import os
import sys
from shutil import rmtree
from typing import Iterable
from typing import Iterator
from typing import NamedTuple
from pathlib import Path
from tempfile import mkdtemp

import sh


def _pydev_fix_fds_after_sh():
    """
    Fix for pycharm's debugger which is trying to monkeypatch stdin/out and that
    is clashing with `sh` module.
    """
    sys.stdout = os.fdopen(1, "w", buffering=1, closefd=False)
    sys.stderr = os.fdopen(2, "w", buffering=1, closefd=False)


if os.getenv("PYCHARM_HOSTED") == "1":
    rdiff_backup = sh.rdiff_backup.bake(_preexec_fn=_pydev_fix_fds_after_sh)
else:
    rdiff_backup = sh.rdiff_backup.bake()


class Increment(NamedTuple):
    timestamp: int

    @staticmethod
    def from_string(line: str) -> "Increment":
        timestamp = int(line.strip().split()[0])
        return Increment(timestamp)


class RdiffAPI:
    def __init__(
        self,
        rsync_dir: str | Path,
        tmp_dir: str | Path | None = None,
        disable_compression: bool = False,
    ) -> None:
        self.rsync_dir = rsync_dir

        self._tmp_dir = tmp_dir
        if self._tmp_dir is None:
            self._tmp_dir = mkdtemp()

        self._disable_compression = disable_compression

        self._options: list[str] = []

    def __del__(self) -> None:
        if os.path.exists(self._tmp_dir):
            rmtree(self._tmp_dir)

    def yield_increments(self) -> Iterator[Increment]:
        increments = rdiff_backup(
            "--parsable-output",
            "-l",
            self.rsync_dir,
        )

        for increment_line in increments.strip().splitlines():
            if increment_line.strip():
                yield Increment.from_string(increment_line)

    def restore(self, out_dir: str | Path, time: int) -> None:
        rdiff_backup("-r", time, "--current-time", time, str(self.rsync_dir), out_dir)

    def add_increment(self, from_dir: str | Path, timestamp: int = 0) -> None:
        options = self._options[:]
        if timestamp != 0:
            options.extend(("--current-time", timestamp))

        if self._disable_compression:
            options.append("--no-compression")

        options.append(from_dir)
        options.append(str(self.rsync_dir))

        rdiff_backup(*options)

    def restore_into(
        self,
        out_dir: str | Path,
        increments_to_keep: Iterable[Increment],
        disable_compression: bool = False,
    ) -> "RdiffAPI":
        out_rsync = RdiffAPI(out_dir, disable_compression=disable_compression)
        for increment in sorted(increments_to_keep, key=lambda x: x.timestamp):
            rmtree(self._tmp_dir)
            os.mkdir(self._tmp_dir)

            print("Restoring", increment.timestamp)

            self.restore(self._tmp_dir, increment.timestamp)
            out_rsync.add_increment(self._tmp_dir, increment.timestamp)

        rmtree(self._tmp_dir)
        return out_rsync



#! /usr/bin/env python3


def get_version(data):
    def all_same(s):
        return all(x == s[0] for x in s)

    def has_digit(s):
        return any(x.isdigit() for x in s)

    data = data.splitlines()
    return list(
        line for line, underline in zip(data, data[1:])
        if (len(line) == len(underline) and
            all_same(underline) and
            has_digit(line) and
            "." in line),
    )[0]



import os
import os.path
import sys
import argparse

import datetime
from collections import defaultdict

from rdiff_trimmer.rdiff_api import RdiffAPI
from rdiff_trimmer.rdiff_api import Increment


def remove_even(rsync_dir, out_dir, disable_compression=False):
    rsync = RdiffAPI(rsync_dir)

    odd_increments = [
        increment
        for cnt, increment in enumerate(rsync.yield_increments())
        if ((cnt + 1) % 2) != 0
    ]

    rsync.restore_into(out_dir, odd_increments, disable_compression=disable_compression)


def keep_one_for_each_month(
    rsync_dir, out_dir, all_from_last_3_months=True, disable_compression=False
):
    rsync = RdiffAPI(rsync_dir)

    def get_month_date(increment):
        date = datetime.datetime.fromtimestamp(increment.timestamp)
        return date.strftime("%Y-%m")

    month_tracker = defaultdict(list)
    for increment in rsync.yield_increments():
        month_tracker[get_month_date(increment)].append(increment)

    last_from_each_month = {increments[-1] for increments in month_tracker.values()}

    if all_from_last_3_months:
        all_months = sorted(month_tracker.keys())

        last_from_each_month.update(month_tracker[all_months[-1]])

        if len(all_months) >= 2:
            last_from_each_month.update(month_tracker[all_months[-2]])
        if len(all_months) >= 3:
            last_from_each_month.update(month_tracker[all_months[-3]])

    rsync.restore_into(
        out_dir, sorted(last_from_each_month), disable_compression=disable_compression
    )


def _check_multiple_parameters(args):
    counter = 0

    if args.list:
        counter += 1
    if args.each_month:
        counter += 1
    if args.remove_even:
        counter += 1

    return counter > 1


def main(args):
    if args.list and not os.path.exists(args.list):
        sys.stderr.write("Increment list file `%s` not found!\n" % args.list)
        sys.exit(1)

    if not os.path.exists(args.rsync_dir):
        sys.stderr.write("rsync directory `%s` not found!\n" % args.rsync_dir)
        sys.exit(1)

    if _check_multiple_parameters(args):
        sys.stderr.write(
            "You can set only one of --keep-increments OR --one-for-each-month"
            " OR --remove-even, not all at once!\n"
        )
        sys.exit(1)

    if args.out_dir is None:
        args.out_dir = os.path.abspath(args.rsync_dir) + "_trimmed"

    if args.remove_even:
        remove_even(args.rsync_dir, args.out_dir, args.disable_compression)
    elif args.each_month:
        keep_one_for_each_month(args.rsync_dir, args.out_dir, args.disable_compression)
    elif args.list:
        with open(args.list) as f:
            increments_to_keep = [
                Increment.from_string(x.strip().split()[0])
                for x in f.read().splitlines()
            ]

        rdiff = RdiffAPI(args.rsync_dir)
        rdiff.restore_into(
            args.out_dir,
            increments_to_keep,
            disable_compression=args.disable_compression,
        )
    else:
        sys.stderr.write("No action selected. Use --help for list.\n")
        sys.exit(1)


def parse_args_and_call_main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-k",
        "--keep-increments",
        dest="list",
        help="Keep only increments listed in this file.",
    )
    parser.add_argument(
        "-o",
        "--one-for-each-month",
        dest="each_month",
        action="store_true",
        help="Keep only one backup for each month.",
    )
    parser.add_argument(
        "-e",
        "--remove-even",
        action="store_true",
        help="Remove even backups. Reduce number of backups to half.",
    )
    parser.add_argument(
        "-d",
        "--disable-compression",
        action="store_true",
        help="Disable default gzip compression used by rdiff.",
    )
    parser.add_argument(
        "rsync_dir", metavar="RSYNC_DIR", help="Path to the rsync directory."
    )
    parser.add_argument(
        "out_dir",
        default=None,
        nargs="?",
        metavar="OUT_DIR",
        help=(
            "Path to the trimmed OUTPUT rsync directory. "
            "Default `{{RSYNC_DIR}}_trimmed`."
        ),
    )

    args = parser.parse_args()
    main(args)


if __name__ == "__main__":
    parse_args_and_call_main()



import os
import sys
import os.path
import argparse

from rdiff_trimmer.rdiff_api import RdiffAPI


def unpack_into(rdiff_path, out_dir_path):
    if not os.path.exists(out_dir_path):
        os.mkdir(out_dir_path)

    api = RdiffAPI(rdiff_path)
    for inc in api.yield_increments():
        inc_path = os.path.join(out_dir_path, str(inc.timestamp))
        if not os.path.exists(inc_path):
            os.mkdir(inc_path)

        api.restore(inc_path, inc.timestamp)
        yield inc.timestamp, inc_path


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-o",
        "--out",
        required=True,
        help="Output directory where the increments will be unpacked.",
    )
    parser.add_argument(
        "in_path",
        metavar="RDIFF_PATH",
        help="Path to the directory with rdiff increments.",
    )
    args = parser.parse_args()

    if not os.path.exists(args.in_path):
        sys.stderr.write("`%s` doesn't exists!\n")
        sys.exit(1)

    for timestamp, out_path in unpack_into(args.in_path, args.out):
        print(out_path, "unpacked")


if __name__ == "__main__":
    main()



import os
import os.path
import shutil

from pytest import fixture

from rdiff_trimmer.trimmer import keep_one_for_each_month

from rdiff_trimmer.rdiff_api import RdiffAPI
from rdiff_trimmer.rdiff_api import Increment


DIR_PATH = os.path.abspath(os.path.dirname(__file__))
DIR_PATH = os.path.join(DIR_PATH, "test_data")


@fixture
def test_data_rdiff_api(tmp_path):
    shutil.copytree(DIR_PATH, tmp_path, dirs_exist_ok=True)
    return RdiffAPI(tmp_path)


@fixture
def tmpdir_rdiff_api(tmp_path):
    tmpdir = tmp_path / "output"
    tmpdir.mkdir()
    return RdiffAPI(tmpdir)


def test_increment_from_string():
    inc_line = "   1524336818 directory   "

    inc = Increment.from_string(inc_line)

    assert int(inc.timestamp) == 1524336818


def test_yield_increments(test_data_rdiff_api):
    increments = list(test_data_rdiff_api.yield_increments())

    assert len(increments) == 4

    assert int(increments[1].timestamp) == 1524342591
    assert int(increments[2].timestamp) == 1524342605


def test_restore_into(test_data_rdiff_api, tmpdir_rdiff_api):
    increments = list(test_data_rdiff_api.yield_increments())

    tmp_dir = tmpdir_rdiff_api.rsync_dir
    test_data_rdiff_api.restore_into(tmp_dir, increments_to_keep=[increments[0]])

    kept_increments = list(tmpdir_rdiff_api.yield_increments())

    assert len(kept_increments) == 1
    assert kept_increments[0].timestamp == increments[0].timestamp

    assert os.path.exists(os.path.join(tmp_dir, "content"))
    assert not os.path.exists(os.path.join(tmp_dir, "and_more_content"))


def test_keep_one_for_each_month(test_data_rdiff_api, tmp_path):
    tmpdir = tmp_path / "monthly"
    keep_one_for_each_month(test_data_rdiff_api.rsync_dir, tmpdir, False)

    trimmed = RdiffAPI(tmpdir)
    kept_increments = list(trimmed.yield_increments())

    assert len(kept_increments) == 2
    assert kept_increments[0].timestamp == 1524342605
    assert kept_increments[1].timestamp == 1770828942
